/*
 * Copyright (C) 2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/update_identity.hh"
#include "src/cfg.hh"
#include "src/impl/lib/create_contact_state_flag_request.hh"
#include "src/impl/lib/unwrap_uuid_from_psql.hh"
#include "src/impl/lib/wrap_uuid_into_psql.hh"
#include "src/impl/exceptions.hh"
#include "src/impl/strong_type.hh"
#include "src/util/util.hh"

#include "liblog/liblog.hh"

#include "libpg/query.hh"
#include "libpg/sql_state/code.hh"

#include <boost/optional.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <iomanip>
#include <sstream>
#include <string>
#include <tuple>
#include <utility>


namespace Fred {
namespace Identity {
namespace Impl {
namespace Lib {

namespace  {

using ContactHistoryUuid = StrongType<boost::uuids::uuid, struct ContactHistoryUuidTag_>;
using IdentityProvider = StrongType<std::string, struct IdentityProviderTag_>;
using Subject = StrongType<std::string, struct SubjectTag_>;
using ContactId = StrongType<std::uint64_t, struct ContactIdTag_>;
using ContactHistoryId = StrongType<std::uint64_t, struct ContactHistoryIdTag_>;
using IdentityId = StrongType<std::uint64_t, struct IdentityIdTag_>;
using ContactHandle = StrongType<std::string, struct ContactHandleTag_>;
using NameChangeProhibited = StrongType<bool, struct NameChangeProhibitedTag_>;
using OrganizationChangeProhibited = StrongType<bool, struct OrganizationChangeProhibitedTag_>;
using IdentChangeProhibited = StrongType<bool, struct IdentChangeProhibitedTag_>;
using PermanentAddressChangeProhibited = StrongType<bool, struct PermanentAddressChangeProhibitedTag_>;
using NameChangeRequested = StrongType<bool, struct NameChangeRequestedTag_>;
using IdentChangeRequested = StrongType<bool, struct IdentChangeRequestedTag_>;
using PermanentAddressChangeRequested = StrongType<bool, struct PermanentAddressChangeRequestedTag_>;
using Name = StrongType<std::string, struct NameTag_>;
using Street1 = StrongType<std::string, struct Street1Tag_>;
using Street2 = StrongType<std::string, struct Street2Tag_>;
using Street3 = StrongType<std::string, struct Street3Tag_>;
using City = StrongType<std::string, struct CityTag_>;
using StateOrProvince = StrongType<std::string, struct StateOrProvinceTag_>;
using PostalCode = StrongType<std::string, struct PostalCodeTag_>;
using CountryCode = StrongType<std::string, struct CountryCodeTag_>;
using Birthdate = StrongType<std::string, struct BirthdateTag_>;
using LogEntryId = StrongType<std::int64_t, struct LogEntryIdTag_>;
using RegistrarHandle = StrongType<std::string, struct RegistrarHandleTag_>;

struct AttachedContactData
{
    ContactUuid contact_uuid;
    ContactId contact_id;
    ContactHandle contact_handle;
    bool name_change_to_prohibit;
    bool organization_change_to_prohibit;
    bool ident_change_to_prohibit;
    bool permanent_address_change_to_prohibit;
    bool name_change_requested;
    bool ident_change_requested;
    bool permanent_address_change_requested;
};

using AttachedContacts = std::vector<AttachedContactData>;

AttachedContacts get_attached_contacts(
        const LibPg::PgTransaction& tx,
        const IdentityProvider& provider,
        const Subject& subject,
        const Optional<Name>& name,
        const Optional<Street1>& street1,
        const Optional<Street2>& street2,
        const Optional<Street3>& street3,
        const Optional<City>& city,
        const Optional<StateOrProvince>& state_or_province,
        const Optional<PostalCode>& postal_code,
        const Optional<CountryCode>& country_code,
        const boost::optional<boost::gregorian::date>& birthdate)
{
    try
    {
        const auto dbres = exec(
                tx,
                LibPg::make_query(
                        "WITH c AS ("
                            "SELECT c.id, "
                                   "COALESCE(LOWER(REGEXP_REPLACE(TRIM(c.name),'\\s\\s+',' ','g')),'') AS name, "
                                   "COALESCE(LOWER(REGEXP_REPLACE(TRIM(c.street1),'\\s\\s+',' ','g')),'') AS street1, "
                                   "COALESCE(LOWER(REGEXP_REPLACE(TRIM(c.street2),'\\s\\s+',' ','g')),'') AS street2, "
                                   "COALESCE(LOWER(REGEXP_REPLACE(TRIM(c.street3),'\\s\\s+',' ','g')),'') AS street3, "
                                   "COALESCE(LOWER(REGEXP_REPLACE(TRIM(c.city),'\\s\\s+',' ','g')),'') AS city, "
                                   "COALESCE(LOWER(REGEXP_REPLACE(TRIM(c.stateorprovince),'\\s\\s+',' ','g')),'') AS stateorprovince, "
                                   "COALESCE(LOWER(REGEXP_REPLACE(TRIM(c.postalcode),'\\s\\s+',' ','g')),'') AS postalcode, "
                                   "COALESCE(LOWER(REGEXP_REPLACE(TRIM(c.country),'\\s\\s+',' ','g')),'') AS country, "
                                   "COALESCE((SELECT TRIM(c.ssn) FROM enum_ssntype WHERE type = 'BIRTHDAY' AND c.ssntype = id),'') AS birthdate, "
                                   "lock_object_state_request_lock(c.id) "
                            "FROM contact_identity ci "
                            "JOIN contact c ON c.id = ci.contact_id "
                            "WHERE ci.identity_provider = ") << LibPg::parameter<IdentityProvider>().as_varchar() << " AND "
                                  "ci.subject = " << LibPg::parameter<Subject>().as_varchar() << " AND "
                                  "ci.valid_to IS NULL "
                            "ORDER BY c.id "
                            "FOR UPDATE) "
                        "SELECT" << LibPg::item<ContactUuid>() << "obr.uuid"
                                 << LibPg::item<ContactId>() << "obr.id"
                                 << LibPg::item<ContactHandle>() << "obr.name"
                                 << LibPg::item<NameChangeProhibited>() << "COALESCE(BOOL_OR("
                                        "eos.name = " << LibPg::parameter<ContactStateFlag::ServerContactNameChangeProhibited>().as_varchar() << "), False)"
                                 << LibPg::item<OrganizationChangeProhibited>() << "COALESCE(BOOL_OR("
                                        "eos.name = " << LibPg::parameter<ContactStateFlag::ServerContactOrganizationChangeProhibited>().as_varchar() << "), False)"
                                 << LibPg::item<IdentChangeProhibited>() << "COALESCE(BOOL_OR("
                                        "eos.name = " << LibPg::parameter<ContactStateFlag::ServerContactIdentChangeProhibited>().as_varchar() << "), False)"
                                 << LibPg::item<PermanentAddressChangeProhibited>() << "COALESCE(BOOL_OR("
                                        "eos.name = " << LibPg::parameter<ContactStateFlag::ServerContactPermanentAddressChangeProhibited>().as_varchar() << "), False)"
                                 << LibPg::item<NameChangeRequested>() << "COALESCE(c.name != LOWER(REGEXP_REPLACE(TRIM(" << LibPg::parameter<Optional<Name>>().as_varchar() << "),'\\s\\s+',' ','g')), False)"
                                 << LibPg::item<PermanentAddressChangeRequested>() << "COALESCE("
                                        "c.street1 != LOWER(REGEXP_REPLACE(TRIM(" << LibPg::parameter<Optional<Street1>>().as_varchar() << "),'\\s\\s+',' ','g')) OR "
                                        "c.street2 != LOWER(REGEXP_REPLACE(TRIM(" << LibPg::parameter<Optional<Street2>>().as_varchar() << "),'\\s\\s+',' ','g')) OR "
                                        "c.street3 != LOWER(REGEXP_REPLACE(TRIM(" << LibPg::parameter<Optional<Street3>>().as_varchar() << "),'\\s\\s+',' ','g')) OR "
                                        "c.city != LOWER(REGEXP_REPLACE(TRIM(" << LibPg::parameter<Optional<City>>().as_varchar() << "),'\\s\\s+',' ','g')) OR "
                                        "c.stateorprovince != LOWER(REGEXP_REPLACE(TRIM(" << LibPg::parameter<Optional<StateOrProvince>>().as_varchar() << "),'\\s\\s+',' ','g')) OR "
                                        "c.postalcode != LOWER(REGEXP_REPLACE(TRIM(" << LibPg::parameter<Optional<PostalCode>>().as_varchar() << "),'\\s\\s+',' ','g')) OR "
                                        "c.country != LOWER(REGEXP_REPLACE(TRIM(" << LibPg::parameter<Optional<CountryCode>>().as_varchar() << "),'\\s\\s+',' ','g')), False)"
                                 << LibPg::item<Birthdate>() << "c.birthdate "
                        "FROM c "
                        "JOIN object_registry obr ON obr.id = c.id "
                        "LEFT JOIN object_state_request osr ON osr.object_id = obr.id AND "
                                                              "osr.valid_from <= NOW() AND "
                                                              "osr.valid_to IS NULL AND "
                                                              "osr.canceled IS NULL AND "
                                                              "osr.state_id IN (SELECT id "
                                                                               "FROM enum_object_states "
                                                                               "WHERE name IN (" <<
                                                                    LibPg::parameter<ContactStateFlag::ServerContactNameChangeProhibited>().as_varchar() << ", " <<
                                                                    LibPg::parameter<ContactStateFlag::ServerContactOrganizationChangeProhibited>().as_varchar() << ", " <<
                                                                    LibPg::parameter<ContactStateFlag::ServerContactIdentChangeProhibited>().as_varchar() << ", " <<
                                                                    LibPg::parameter<ContactStateFlag::ServerContactPermanentAddressChangeProhibited>().as_varchar() << ")) "
                        "LEFT JOIN enum_object_states eos ON eos.id = osr.state_id "
                        "GROUP BY obr.id,"
                                 "c.id,"
                                 "c.name,"
                                 "c.street1,"
                                 "c.street2,"
                                 "c.street3,"
                                 "c.city,"
                                 "c.stateorprovince,"
                                 "c.postalcode,"
                                 "c.country,"
                                 "c.birthdate "
                        "ORDER BY obr.id",
                {provider,
                 subject,
                 name,
                 street1,
                 street2,
                 street3,
                 city,
                 state_or_province,
                 postal_code,
                 country_code,
                 ContactStateFlag::ServerContactNameChangeProhibited{},
                 ContactStateFlag::ServerContactOrganizationChangeProhibited{},
                 ContactStateFlag::ServerContactIdentChangeProhibited{},
                 ContactStateFlag::ServerContactPermanentAddressChangeProhibited{}});
        AttachedContacts result;
        for (auto&& row : dbres)
        {
            AttachedContactData contact;
            contact.contact_uuid = row.get<ContactUuid>();
            contact.contact_id = row.get<ContactId>();
            contact.contact_handle = row.get<ContactHandle>();
            contact.name_change_to_prohibit = !*row.get<NameChangeProhibited>() && (name != Name::nullopt);
            contact.organization_change_to_prohibit = !*row.get<OrganizationChangeProhibited>();
            contact.ident_change_to_prohibit = !*row.get<IdentChangeProhibited>() && (birthdate != boost::none);
            contact.permanent_address_change_to_prohibit = !*row.get<PermanentAddressChangeProhibited>() &&
                    (street1 != Street1::nullopt ||
                     street2 != Street2::nullopt ||
                     street3 != Street3::nullopt ||
                     city != City::nullopt ||
                     state_or_province != StateOrProvince::nullopt ||
                     postal_code != PostalCode::nullopt ||
                     country_code != CountryCode::nullopt);
            contact.name_change_requested = *row.get<NameChangeRequested>();
            contact.ident_change_requested = [&]()
            {
                if (birthdate == boost::none)
                {
                    return false;
                }
                const auto current_birthdate = *row.get<Birthdate>();
                if (current_birthdate.empty())
                {
                    return true;
                }
                try
                {
                    const auto current_birthdate_value = Util::birthdate_from_string_to_date(current_birthdate);
                    return current_birthdate_value != *birthdate;
                }
                catch (...)
                {
                    return true;
                }
            }();
            contact.permanent_address_change_requested = *row.get<PermanentAddressChangeRequested>();
            result.push_back(std::move(contact));
        }
        return result;
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("exception caught: {}", e.what());
        throw;
    }
    catch (...)
    {
        LIBLOG_ERROR("unexpected exception caught");
        throw;
    }
}

boost::gregorian::date to_boost_date(const Unwrapper::Date& date)
{
    return boost::gregorian::date{boost::gregorian::date::ymd_type{
            static_cast<short unsigned>(date.get_year()),
            static_cast<short unsigned>(date.get_month()),
            static_cast<short unsigned>(date.get_day())}};
}

std::string to_string(const Unwrapper::Date& date)
{
    std::ostringstream out;
    out << date.get_year() << "-"
        << std::setw(2) << std::setfill('0') << date.get_month() << "-"
        << std::setw(2) << std::setfill('0') << date.get_day();
    return out.str();
}

std::string make_what_to_set_part(const Unwrapper::IdentityData& data, LibPg::ExtendableQueryArguments& arguments)
{
    std::string to_set;
    if (data.has_name())
    {
        to_set = "name = " + push_back_argument(arguments, data.get_name()).as_varchar();
    }
    if (data.has_address())
    {
        if (!to_set.empty())
        {
            to_set += ", ";
        }
        const auto addr = data.get_address();
        const auto street = addr.get_street();
        to_set += "street1 = " + push_back_argument(arguments, street[0]).as_varchar() + ", "
                  "street2 = " + push_back_argument(arguments, street[1]).as_varchar() + ", "
                  "street3 = " + push_back_argument(arguments, street[2]).as_varchar() + ", "
                  "city = " + push_back_argument(arguments, addr.get_city()).as_varchar() + ", "
                  "stateorprovince = " + push_back_argument(arguments, addr.get_state_or_province()).as_varchar() + ", "
                  "postalcode = " + push_back_argument(arguments, addr.get_postal_code()).as_varchar() + ", "
                  "country = " + push_back_argument(arguments, addr.get_country_code()).as_varchar();
    }
    if (data.has_birthdate())
    {
        if (!to_set.empty())
        {
            to_set += ", ";
        }
        to_set += "ssntype = (SELECT id FROM enum_ssntype WHERE type = 'BIRTHDAY'::TEXT), "
                  "ssn = " + push_back_argument(arguments, to_string(data.get_birthdate())).as_varchar();
    }
    return to_set;
}

ContactHistoryId update_contact(
        const LibPg::PgRwTransaction& tx,
        const ContactId& contact_id,
        const Unwrapper::IdentityData& data,
        const Optional<RegistrarHandle>& update_by_registrar,
        const Optional<LogEntryId>& log_entry_id)
{
    LibPg::ExtendableQueryArguments arguments;
    const auto sql =
            "WITH "
            "new_history AS ("
                "INSERT INTO history(request_id) "
                "VALUES (" + push_back_argument(arguments, log_entry_id).as_big_int() + ") "
                "RETURNING id), "
            "copy_object_history AS ("
                "INSERT INTO object_history(historyid, id, clid, upid, trdate, update, authinfopw) "
                "SELECT h.id, o.id, o.clid, o.upid, o.trdate, o.update, o.authinfopw "
                "FROM object o, "
                     "new_history h "
                "WHERE o.id = " + push_back_argument(arguments, contact_id).as_big_int() + " "
                "RETURNING historyid AS history_id, id AS contact_id), "
            "update_object AS ("
                "UPDATE object o "
                "SET update = NOW(), "
                    "upid = (SELECT id FROM registrar "
                            "WHERE UPPER(handle) = "
                                  "UPPER(" + push_back_argument(arguments, update_by_registrar).as_varchar() + ")) "
                "FROM copy_object_history coh "
                "WHERE o.id = coh.contact_id "
                "RETURNING coh.history_id, coh.contact_id), "
            "update_object_registry AS ("
                "UPDATE object_registry obr "
                "SET historyid = uo.history_id "
                "FROM update_object uo "
                "WHERE obr.id = uo.contact_id "
                "RETURNING uo.history_id, uo.contact_id) "
            "UPDATE contact c "
            "SET " + make_what_to_set_part(data, arguments) + " "
            "FROM update_object_registry uor "
            "WHERE c.id = uor.contact_id "
            "RETURNING uor.history_id";
    const auto db_res = LibPg::PgResultTuples{exec(tx, sql, arguments)};
    if (db_res.empty())
    {
        struct ContactDoesNotExist : Exception
        {
            const char* what() const noexcept override { return "contact to update not found"; }
        };
        throw ContactDoesNotExist{};
    }
    if (db_res.size() != LibPg::RowIndex{1})
    {
        struct TooManyContacts : Exception
        {
            const char* what() const noexcept override { return "just one contact to update expected"; }
        };
        throw TooManyContacts{};
    }
    const auto contact_history_id = db_res[LibPg::RowIndex{0}][LibPg::ColumnIndex{0}].as<ContactHistoryId>();
    const auto copy_contact_result = exec(
            tx,
            LibPg::make_query(
                    "WITH "
                    "copy_contact_address_history AS ("
                        "INSERT INTO contact_address_history ("
                            "historyid, id, contactid, type, company_name, street1, street2, street3, city, "
                            "stateorprovince, postalcode, country) "
                        "SELECT ") << LibPg::parameter<ContactHistoryId>().as_big_int() << ", "
                            "id, contactid, type, company_name, street1, street2, street3, city, stateorprovince, "
                            "postalcode, country "
                        "FROM contact_address "
                        "WHERE contactid = " << LibPg::parameter<ContactId>().as_big_int() << ") "
                    "INSERT INTO contact_history ("
                        "historyid, id, name, organization, street1, street2, street3, city, stateorprovince, "
                        "postalcode, country, telephone, fax, email, notifyemail, vat, ssntype, ssn, disclosename, "
                        "discloseorganization, discloseaddress, disclosetelephone, disclosefax, discloseemail, "
                        "disclosevat, discloseident, disclosenotifyemail, warning_letter) "
                    "SELECT " << LibPg::parameter<ContactHistoryId>().as_big_int() << ", "
                        "id, name, organization, street1, street2, street3, city, stateorprovince, postalcode, "
                        "country, telephone, fax, email, notifyemail, vat, ssntype, ssn, disclosename, "
                        "discloseorganization, discloseaddress, disclosetelephone, disclosefax, discloseemail, "
                        "disclosevat, discloseident, disclosenotifyemail, warning_letter "
                    "FROM contact "
                    "WHERE id = " << LibPg::parameter<ContactId>().as_big_int(),
            {contact_id, contact_history_id});
    if (copy_contact_result.get_number_of_rows_affected() < LibPg::RowIndex{1})
    {
        struct ContactDoesNotExist : Exception
        {
            const char* what() const noexcept override { return "there are no contact data to copy"; }
        };
        throw ContactDoesNotExist{};
    }
    return contact_history_id;
}

}//namespace Fred::Identity::Impl::Lib::{anonymous}

void update_identity(const LibPg::PgRwTransaction& tx, const Unwrapper::UpdateIdentityRequest& request)
{
    const auto streets = request.get_identity_data().has_address() ? request.get_identity_data().get_address().get_street()
                                                                   : std::vector<std::string>{};
    const auto has_street = [&](std::size_t idx) { return idx < (streets.size() + 1); };
    const auto attached_contacts = get_attached_contacts(
            tx,
            IdentityProvider{request.get_identity().get_identity_provider().get_value()},
            Subject{request.get_identity().get_subject().get_value()},
            Optional<Name>{request.get_identity_data().has_name(), [&]() { return Name{request.get_identity_data().get_name()}; }},
            Optional<Street1>{has_street(1), [&]() { return Street1{streets[0]}; }},
            Optional<Street2>{has_street(2), [&]() { return Street2{streets[1]}; }},
            Optional<Street3>{has_street(3), [&]() { return Street3{streets[2]}; }},
            Optional<City>{request.get_identity_data().has_address(), [&]() { return City{request.get_identity_data().get_address().get_city()}; }},
            Optional<StateOrProvince>{request.get_identity_data().has_address(), [&]() { return StateOrProvince{request.get_identity_data().get_address().get_state_or_province()}; }},
            Optional<PostalCode>{request.get_identity_data().has_address(), [&]() { return PostalCode{request.get_identity_data().get_address().get_postal_code()}; }},
            Optional<CountryCode>{request.get_identity_data().has_address(), [&]() { return CountryCode{request.get_identity_data().get_address().get_country_code()}; }},
            request.get_identity_data().has_birthdate() ? to_boost_date(request.get_identity_data().get_birthdate()) : boost::optional<boost::gregorian::date>{});
    const auto log_entry_id = Optional<LogEntryId>{
            request.has_log_entry_id(),
            [&]() { return LogEntryId{request.get_log_entry_id().get_value()}; }};
    const auto registrar = Optional<RegistrarHandle>{
            Cfg::Options::get().fred.registrar_originator != boost::none,
            [&]() { return RegistrarHandle{*Cfg::Options::get().fred.registrar_originator}; }};
    std::for_each(begin(attached_contacts), end(attached_contacts), [&](auto&& contact)
            {
                if (contact.name_change_to_prohibit)
                {
                    create_contact_state_flag_request(tx, contact.contact_uuid, ContactStateFlag::ServerContactNameChangeProhibited{});
                }
                if (contact.organization_change_to_prohibit)
                {
                    create_contact_state_flag_request(tx, contact.contact_uuid, ContactStateFlag::ServerContactOrganizationChangeProhibited{});
                }
                if (contact.ident_change_to_prohibit)
                {
                    create_contact_state_flag_request(tx, contact.contact_uuid, ContactStateFlag::ServerContactIdentChangeProhibited{});
                }
                if (contact.permanent_address_change_to_prohibit)
                {
                    create_contact_state_flag_request(tx, contact.contact_uuid, ContactStateFlag::ServerContactPermanentAddressChangeProhibited{});
                }
                if (contact.name_change_requested ||
                    contact.ident_change_requested ||
                    contact.permanent_address_change_requested)
                {
                    update_contact(
                            tx,
                            contact.contact_id,
                            request.get_identity_data(),
                            registrar,
                            log_entry_id);
                }
            });
}

}//namespace Fred::Identity::Impl::Lib
}//namespace Fred::Identity::Impl
}//namespace Fred::Identity
}//namespace Fred
